import json
import aiomysql
from datetime import datetime, date
from tracardi_dot_notation.dict_traverser import DictTraverser
from tracardi.service.storage.driver import storage
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData, Form, FormGroup, FormField, FormComponent
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result
from tracardi_mysql_connector.model.configuration import Configuration
from tracardi_mysql_connector.model.connection import Connection


def validate(config: dict) -> Configuration:
    return Configuration(**config)


class MysqlConnectorAction(ActionRunner):

    @staticmethod
    async def build(**kwargs) -> 'MysqlConnectorAction':

        configuration = validate(kwargs)
        source = await storage.driver.resource.load(configuration.source.id)
        connection = Connection(**source.config)
        pool = await connection.connect(configuration.timeout)

        return MysqlConnectorAction(pool, configuration)

    def __init__(self, pool, config: Configuration):
        self.pool = pool
        self.config = config

    async def run(self, payload):

        # Prepare statement data
        template = DictTraverser(self._get_dot_accessor(payload))
        data = template.reshape(self.config.data)
        self.console.log("Executing query: {} with data: {}".format(self.config.query, data))

        async with self.pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cursor:
                if self.config.type == 'call':
                    # todo implement

                    return Result(port="result", value={"result": payload})
                else:
                    if len(data) > 0:
                        await cursor.execute(self.config.query, tuple(data))
                    else:
                        await cursor.execute(self.config.query)

                    if self.config.type in ['insert', 'delete', 'update']:
                        await conn.commit()
                        if self.config.type == 'insert':
                            return Result(port="result", value={"last_insert_id": cursor.lastrowid})
                    if self.config.type == 'select':
                        result = await cursor.fetchall()
                        result = [self.to_dict(record) for record in result]
                        return Result(port="result", value={"result": result})

                return Result(port="result", value=payload)

    async def close(self):
        if self.pool:
            self.pool.close()
            await self.pool.wait_closed()

    async def on_error(self):
        await self.close()

    @staticmethod
    def to_dict(record):

        def json_default(obj):
            """JSON serializer for objects not serializable by default json code"""

            if isinstance(obj, (datetime, date)):
                return obj.isoformat()

            return obj.decode('utf-8')

        j = json.dumps(dict(record), default=json_default)
        return json.loads(j)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi_mysql_connector.plugin',
            className='MysqlConnectorAction',
            inputs=["payload"],
            outputs=['result'],
            version='0.6.0',
            license="MIT",
            author="Risto Kowaczewski",
            manual="mysql_connector_action",
            init={
                "source": {
                    "id": "",
                    "name": ""
                },
                "type": "select",
                "query": "SELECT 1",
                "data": [],
                "timeout": 10
            },
            form=Form(groups=[
                FormGroup(
                    name="MySQL Resource",
                    fields=[
                        FormField(
                            id="source",
                            name="MySQL resource",
                            description="Select MySQL resource. Credentials from selected resource will be used "
                                        "to connect to database.",
                            component=FormComponent(type="resource", props={"label": "resource"})
                        )
                    ]
                ),
                FormGroup(
                    name="MySQL Query",
                    fields=[
                        FormField(
                            id="type",
                            name="SQL type",
                            description="Select SQL query type. It must match typed SQL statement otherwise it may not "
                                        "execute properly.",
                            component=FormComponent(
                                type="select",
                                props={
                                    "label": "SQL type",
                                    "items": {
                                        "select": "SELECT",
                                        "insert": "INSERT",
                                        "delete": "DELETE",
                                        "create": "CREATE",
                                        "update": "UPDATE",
                                    }
                                })
                        ),
                        FormField(
                            id="query",
                            name="SQL statement",
                            description="Type prepared SQL query to be executed in MySQL. Use `%s` for data "
                                        "placeholders. Below tape data to replace placeholder.",
                            component=FormComponent(
                                type="sql",
                                props={
                                    "label": "SQL"
                                })
                        ),
                        FormField(
                            id="data",
                            name="Data for SQL statement",
                            description="Type data for prepared SQL statement. Data must be a list "
                                        "of values or paths to values. It will replace %s placeholders in SQL statment. "
                                        "It will replace placeholders one by one so order matters.",
                            component=FormComponent(
                                type="listOfDotPaths",
                                props={"label": "List of data"})
                        )
                    ]
                ),
                FormGroup(
                    name="Advanced MySQL Configuration",
                    fields=[
                        FormField(
                            id="timeout",
                            name="Connection time-out",
                            description="Type SQL query connection time-out",
                            component=FormComponent(
                                type="text",
                                props={
                                    "label": "Time-out"
                                })
                        )
                    ]
                ),
            ]),

        ),
        metadata=MetaData(
            name='MySQL connector',
            desc='Connects to mysql and reads data.',
            type='flowNode',
            width=200,
            height=100,
            icon='mysql',
            group=["Connectors"]
        )
    )
